/*

<GotohDPdis.c>

Functions for Pairwise Sequence Alignment
based on "Waterman-Smith-Beyer" and "Gotoh"

coded by Takeshi Kawabata

REFERENCE

Gotoh.O "An improved algorithm for matching biological sequences"
J.Mol.Biol. (1982) 162 705-708

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "main.h"
#include "protein.h"
#include "ScoreENV.h"
#include "ScoreDIS.h"
#include "ScoreSSE.h"
#include "LlistAlign.h"


/*
 -----j
|
|
i

*/


/****** Variable for Forward *******/
static float **D;      /* [Naa+1][Naa+1] (malloc later) : Node Score for Diagonal   */ 
static float **P;      /* [Naa+1][Naa+1] (malloc later) : Node Score for Vertical   */ 
static float **Q;      /* [Naa+1][Naa+1] (malloc later) : Node Score for Horizontal */  

static char  **path;   /* [Naa+2][Naa+2] (malloc later) : 'd'iagonal,'h'orizontal,'v'ertical */  
static char  **vpath;  /* [Naa+2][Naa+2] (malloc later) : 'o'pen,'e'xtend  */  
static char  **hpath;  /* [Naa+2][Naa+2] (malloc later) : 'o'pen,'e'xtend  */  
static char  **mark;   /* [Naa+2][Naa+2] (malloc later) :  0:ignore 1:focus */  

/***** AlignNode for Iterative GotohDP ****/
static struct AlignNode PRE;

/******** Boundary Condition **********/
static char etypeA,etypeB;  /* If '0', set end   penalty = 0  */ 
static char btypeA,btypeB;  /* If '0', set begin penalty = 0  */ 
static char Itype;

static float InfMin;

/******** FUNCTIONS (GLOBAL) ********/
float GotohDP();
float GotohDP_Dscore();
float GotohDP_Dscore_Iterative();
void   Malloc_GotohDP_Matrix();
float Calculate_AlignScore();

/******** FUNCTIONS (LOCAL) ********/
static void Set_Initial_Path();
static void Set_All_Path();
static void Trace_Path();
static void Find_Optimal_Local_Node();
static void Trace_Path_Both_Direction();
static void Set_All_Path_Dscore();




void Malloc_GotohDP_Matrix(n,m)
 int n,m;
{
  int i,N,M;

  /* D,P,Q need [0..NaaA  ][0..NaaB  ]  */
  /* paths need [0..NaaA+1][0..NaaB+1]  */

  N = n + 2; M = m + 2;
  
  D   = (float **) malloc(sizeof(float *) * N);
  P   = (float **) malloc(sizeof(float *) * N);
  Q   = (float **) malloc(sizeof(float *) * N);
 
  
  path  = (char **) malloc(sizeof(char *) * N);
  vpath = (char **) malloc(sizeof(char *) * N);
  hpath = (char **) malloc(sizeof(char *) * N);
  mark  = (char **) malloc(sizeof(char *) * N);
  
 
  for (i=0;i<N;++i)
   {
    D[i]     = (float *) malloc(sizeof(float)* M);
    P[i]     = (float *) malloc(sizeof(float)* M);
    Q[i]     = (float *) malloc(sizeof(float)* M);
   
    path[i]  = (char *) malloc(sizeof(char)* M);
    vpath[i] = (char *) malloc(sizeof(char)* M);
    hpath[i] = (char *) malloc(sizeof(char)* M);
    mark[i]  = (char *) malloc(sizeof(char)* M);
    }


  Malloc_AlignNode(&PRE,N+M,'-');
 
  InfMin = -100000.0; 

} /* end of Malloc_GotohDP_Matrix() */





float GotohDP(proA,proB,an,sn)
 struct PROTEIN *proA,*proB;
 struct AlignNode *an;
 struct AlignNode *sn; /** SSE constrant Alignment ***/
{
  int i,j,si,sj; 
  float sc;

  Itype = PAR.AlgType;   
  
       if (Itype == 'B'){ etypeA = etypeB = btypeA = btypeB = '0';}  
  else if (Itype == 'M')
    { if (proA->Naa > proB->Naa) { etypeA = btypeA = '0'; etypeB = btypeB = '-';}  
      if (proB->Naa > proA->Naa) { etypeA = btypeA = '-'; etypeB = btypeB = '0';}  }
  else if (Itype == 'L')  etypeA = etypeB = '0';
  else etypeA = etypeB = btypeA = btypeB = '-';

  Set_Initial_Path(proA,proB,PAR.GapExtE,PAR.GapOpenE);
  Set_All_Path(proA,proB,sn);
  if (Itype=='L') Find_Optimal_Local_Node(proA,proB,&si,&sj,&sc);
  else { si = 1; sj = 1; }            

  Trace_Path(proA,proB,si,sj,an); 

  if (Itype=='L') return(sc); else return(D[0][0]); 

} /* end of GotohDP() */



float GotohDP_Dscore_Iterative(proA,proB,an,DmapA,DmapB,sn)
 struct PROTEIN *proA,*proB;
 struct AlignNode *an; 
 struct MAP2D *DmapA, *DmapB; 
 struct AlignNode *sn;  /** SSE constraint alignment **/ 
{ 
  double ScEnv,ScDis,ScPre,ScPrePre,ScBest,eps;
  int i,end;

  eps = 0.01;
  ScDis =   ScoreDistanceS(proA,proB,an,DmapA,DmapB)
            + (float)GapPenalty(an,PAR.GapExtD,PAR.GapOpenD);
  
  ScBest = ScDis; 
  ScPre = ScPrePre = InfMin; 
  if (PAR.OtypeD == 'T') printf("#Iterative_Align : INIT ScDis %f\n",ScDis); 
  i = end = 0;

  /***
  [Criteria to stop iteration]
   (1) Sc and ScPre    is the same, and Sc is the best score among the paths already appeared.
   (2) Sc and ScPrePre is the same, and Sc is the best score among the paths already appeared.
   (3) i > Nrepeat, and Sc >= ScPre.
        ( (2) and (3) are for 2-period fructuation
          up -> down -> up -> down ->..., we want to stop 'up' path.)   
   (4) i > Nrepeat *1.5 
  ***/
 
  while (end==0)
  { 
    ScPrePre = ScPre; 
    ScPre    = ScDis; 
  
  
    ScEnv = GotohDP_Dscore(proA,proB,an,DmapA,DmapB,sn);
    ScEnv -= (float)AllSseOffsetSc(proA,proB,an,sn,'D'); 
    ScDis = ScoreDistanceS(proA,proB,an,DmapA,DmapB)
            + (float)GapPenalty(an,PAR.GapExtD,PAR.GapOpenD);

    if (PAR.OtypeD == 'T') 
        printf("#Iteration : [%d] ScEnv %f ScDis %f Pre %f PPre %f Best %f Nali %d\n",
                                    i,ScEnv,ScDis,ScPre,ScPrePre,ScBest,an->Nali); 
    if ((fabs(ScDis-ScPre)<eps)   &&(fabs(ScDis-ScBest)<eps)) end = 1; 
    if ((fabs(ScDis-ScPrePre)<eps)&&(fabs(ScDis-ScBest)<eps)) end = 1; 
    if ((i>=PAR.Nrepeat)&&(ScDis >= ScPre)) end = 1; 
    if (i >= 1.5*PAR.Nrepeat) end = 1;
    if (ScDis > ScBest) ScBest = ScDis; 
    ++i;
  }

  ScoreDistanceSPair(proA,proB,an,DmapA,DmapB);

  if (PAR.OtypeD =='T') printf("#Iterative_Align : Converge count [%d]\n",i);
  return(ScDis);

} /* end of GotohDP_Dscore_Iterative() */





float GotohDP_Dscore(proA,proB,an,DmapA,DmapB,sn)
 struct PROTEIN *proA,*proB;
 struct AlignNode *an;
 struct MAP2D *DmapA, *DmapB;
 struct AlignNode *sn;  /** Constraint for SSE Alignment **/
{
  int i,j,p,q,k,si,sj,max_i,min_i,max_j,min_j; 
  float sc;

  if (PRE.MaxNali < an->Nali)
  { printf("#ERROR:PRE.MaxNali %s is less than Nali %d\n",PRE.MaxNali,an->Nali);
    exit(1); }

  PRE.Nali = an->Nali;  
  for (i=1;i<=PRE.Nali;++i) 
   { PRE.nalA[i] = an->nalA[i]; PRE.nalB[i] = an->nalB[i]; }

  Itype = PAR.AlgType;   
  
       if (Itype == 'B'){ etypeA = etypeB = btypeA = btypeB = '0';}  
  else if (Itype == 'M')
    { if (proA->Naa > proB->Naa) { etypeA = btypeA = '0'; etypeB = btypeB = '-';}  
      if (proB->Naa > proA->Naa) { etypeA = btypeA = '-'; etypeB = btypeB = '0';}  }
  else if (Itype == 'L')  etypeA = etypeB = '0';
  else etypeA = etypeB = btypeA = btypeB = '-';

  /**** Making Initial Alignment Neighbor Map (mark:1 neighbor) ****/
 
  for (i=0;i<=proA->Naa;++i)
   for (j=0;j<=proB->Naa;++j) mark[i][j] = 0; 

  i = j = 0;

  for (k=1;k<= PRE.Nali;++k)
    {
     if (PRE.nalA[k]>0) i = PRE.nalA[k];   
     if (PRE.nalB[k]>0) j = PRE.nalB[k];   
 
     min_i = i - PAR.Wmark; max_i = i + PAR.Wmark; 
     min_j = j - PAR.Wmark; max_j = j + PAR.Wmark; 
     if (min_i<0) min_i = 0; 
     if (min_j<0) min_j = 0; 
     if (max_i >proA->Naa) max_i = proA->Naa; 
     if (max_j >proB->Naa) max_j = proB->Naa; 
     for (p=min_i;p<=max_i;++p) 
      for (q=min_j;q<=max_j;++q) mark[p][q] = 1; 
   }
  
  Set_Initial_Path(proA,proB,PAR.GapExtD,PAR.GapOpenD);
  Set_All_Path_Dscore(proA,proB,&PRE,DmapA,DmapB,sn);
  if (Itype=='L') Find_Optimal_Local_Node(proA,proB,&si,&sj,&sc);
  else { si = 1; sj = 1; }            
  

  Trace_Path(proA,proB,si,sj,an); 
  
  /* printf("sc %f si %d sj %d\n",sc,si,sj); */

  if (Itype=='L') return(sc); else return(D[0][0]); 

} /* end of GotohDP_Dscore() */




void Find_Optimal_Local_Node(proA,proB,si,sj,sc)
 struct PROTEIN *proA,*proB;
 int *si,*sj;
 float *sc;
{ int i,j,max_i,max_j;
  float max;

  max_i = 0; max_j = 0; max = D[0][0];
 
  for (i=0;i<=proA->Naa;++i)
   for (j=0;j<=proB->Naa;++j)
     if (D[i][j]>max) {max = D[i][j]; max_i = i; max_j = j;} 
 
  *si = max_i+1;  *sj = max_j+1;
  *sc = max;


}  /* end of Find_Optimal_Local_Node() */




void Set_Initial_Path(proA,proB,GapExt,GapOpen)
 struct PROTEIN *proA,*proB;
 float GapExt,GapOpen;
{
 int i,j;
 
  for (i=proA->Naa;i>=0;--i)
   { path[i][proB->Naa+1] = 'v'; hpath[i][proB->Naa+1] = 'o'; vpath[i][proB->Naa+1] = 'o'; 
   if (Itype=='L') path[i][proB->Naa+1] = '0'; 
   if (etypeA=='0')  D[i][proB->Naa] = P[i][proB->Naa] = 0.0;
   else 
   {
    if (i==proA->Naa) D[i][proB->Naa] = P[i][proB->Naa] = 0.0;
                else  D[i][proB->Naa] = P[i][proB->Naa] = GapExt * (proA->Naa-i) + GapOpen;
   
    }
   Q[i][proB->Naa] = InfMin; } 

 for (j=proB->Naa;j>=0;--j)
  { path[proA->Naa+1][j] = 'h'; vpath[proA->Naa+1][j] = 'o'; hpath[proA->Naa+1][j] = 'o'; 
    if (Itype=='L') path[proA->Naa+1][j] = '0'; 
    if (etypeB=='0') D[proA->Naa][j] = Q[proA->Naa][j] = 0.0;
    else
    { 
     if (j==proB->Naa) D[proA->Naa][j] = Q[proA->Naa][j] = 0.0;
                  else D[proA->Naa][j] = Q[proA->Naa][j] = GapExt * (proB->Naa-j-1) + GapOpen;
     } 
    P[proA->Naa][j] = InfMin; } 

 hpath[proA->Naa+1][proB->Naa+1] =  vpath[proA->Naa+1][proB->Naa+1] = 'o'; 
 path[proA->Naa+1][proB->Naa+1] = '0'; 

 D[proA->Naa][proB->Naa] = P[proA->Naa][proB->Naa] = Q[proA->Naa][proB->Naa] = 0.0;

} /* end of Set_Initial_Path() */



void Set_All_Path(proA,proB,sn)
 struct PROTEIN *proA,*proB; 
 struct AlignNode *sn; 
{
 int i,j;
 float h_cost,v_cost,d_cost;
 float Ho_cost,He_cost,Vo_cost,Ve_cost;

 
 for (i=proA->Naa;i>=1;--i)
  for (j=proB->Naa;j>=1;--j)
  {

   /**CAUTION!! This Part is not mathematically strict **/
   /*     it prefer  d>h>v  unreasonably                     */
   /* and it  prefer e>o    unreasonably                     */
   /*****************************************************/

   /* Set P */
   Ve_cost = P[i][j-1] + PAR.GapExtE; 
   Vo_cost = D[i][j-1] + PAR.GapOpenE;

   if ((j==1)&&(btypeA=='0'))
    { Ve_cost = P[i][j-1]; Vo_cost = D[i][j-1]; }
   if (Ve_cost>=Vo_cost) {P[i-1][j-1] = Ve_cost; vpath[i][j] = 'e'; }
                 else    {P[i-1][j-1] = Vo_cost; vpath[i][j] = 'o'; }
   /* Set Q */
   He_cost = Q[i-1][j] + PAR.GapExtE; 
   Ho_cost = D[i-1][j] + PAR.GapOpenE;

   if ((i==1)&&(btypeB=='0'))
    { He_cost = Q[i-1][j]; Ho_cost = D[i-1][j]; }

   if (He_cost>=Ho_cost) {Q[i-1][j-1] = He_cost; hpath[i][j] = 'e'; }
                   else  {Q[i-1][j-1] = Ho_cost; hpath[i][j] = 'o'; }

   /* Set D */
   v_cost = P[i-1][j-1];   
   h_cost = Q[i-1][j-1];  
   d_cost = D[i][j] + ResMatchScore(proA,i,proB,j,sn,'O'); 

       if ((Itype=='L')&&(0.0>=d_cost)&&(0.0>=h_cost)&&(0.0>=v_cost)) 
          { D[i-1][j-1] = 0.0; path[i][j] = '0'; }
  else if ((d_cost>=h_cost)&&(d_cost>=v_cost)) { D[i-1][j-1] = d_cost; path[i][j] = 'd'; }
  else if ((v_cost>=h_cost)&&(v_cost>=d_cost)) { D[i-1][j-1] = v_cost; path[i][j] = 'v'; }
  else if ((h_cost>=v_cost)&&(h_cost>=d_cost)) { D[i-1][j-1] = h_cost; path[i][j] = 'h'; }  
 
  }  


} /* end of Set_All_Path() */



void Set_All_Path_Dscore(proA,proB,pre,DmapA,DmapB,sn)
 struct PROTEIN *proA,*proB; 
 struct AlignNode *pre;      /* previous alignment */ 
 struct MAP2D *DmapA, *DmapB;
 struct AlignNode *sn; 
{
 int i,j;
 float h_cost,v_cost,d_cost;
 float Ho_cost,He_cost,Vo_cost,Ve_cost;
 char   Dmark,Vmark,Hmark;

 for (i=proA->Naa;i>=1;--i)
  for (j=proB->Naa;j>=1;--j)
  {

   /**CAUTION!! This Part is not mathematically strict **/
   /*     it prefer  d>h>v  unreasonably                     */
   /* and it  prefer e>o    unreasonably                     */
   /*****************************************************/

   /** Set Mark **/
   /*    j-1  j 
     i-1      X
              -
     i    X | X
   */ 

   Dmark = Vmark = Hmark = 0;
   if (mark[i][j]   == 1) Dmark = Hmark = Vmark = 1;
   if (mark[i-1][j] == 1) Hmark = 1;
   if (mark[i][j-1] == 1) Vmark = 1;

   /** Set P **/
   if (Vmark==1)
   { Ve_cost = P[i][j-1] + PAR.GapExtD; 
     Vo_cost = D[i][j-1] + PAR.GapOpenD; 
     if ((j==1)&&(btypeA=='0'))
      { Ve_cost = P[i][j-1]; Vo_cost = D[i][j-1]; }
     } 
   else Ve_cost = Vo_cost = InfMin;

   if (Ve_cost>=Vo_cost) {P[i-1][j-1] = Ve_cost; vpath[i][j] = 'e'; }
                 else    {P[i-1][j-1] = Vo_cost; vpath[i][j] = 'o'; }
   /** Set Q **/
   if (Hmark==1)
    { He_cost = Q[i-1][j] + PAR.GapExtD; 
      Ho_cost = D[i-1][j] + PAR.GapOpenD; 
      if ((i==1)&&(btypeB=='0'))
       { He_cost = Q[i-1][j]; Ho_cost = D[i-1][j]; }
     }
   else He_cost = Ho_cost = InfMin;
    
   if (He_cost>=Ho_cost) {Q[i-1][j-1] = He_cost; hpath[i][j] = 'e'; }
                   else  {Q[i-1][j-1] = Ho_cost; hpath[i][j] = 'o'; }

   /** Set D **/
   v_cost = P[i-1][j-1];   
   h_cost = Q[i-1][j-1];  
 
   if (Dmark == 1) 
    d_cost = D[i][j] + DisScore_with_PreAli(proA,proB,i,j,pre,DmapA,DmapB,sn); 
   else d_cost = InfMin;  
 
       if ((Itype=='L')&&(0.0>=d_cost)&&(0.0>=h_cost)&&(0.0>=v_cost)) 
          { D[i-1][j-1] = 0.0; path[i][j] = '0'; }
  else if ((d_cost>=h_cost)&&(d_cost>=v_cost)) { D[i-1][j-1] = d_cost; path[i][j] = 'd'; }
  else if ((v_cost>=h_cost)&&(v_cost>=d_cost)) { D[i-1][j-1] = v_cost; path[i][j] = 'v'; }
  else if ((h_cost>=v_cost)&&(h_cost>=d_cost)) { D[i-1][j-1] = h_cost; path[i][j] = 'h'; }  
  }  

} /* end of Set_All_Path_Dscore() */




void Trace_Path(proA,proB,si,sj,a)
 struct PROTEIN *proA,*proB;
 int si,sj;
 struct AlignNode *a;
{
 int i,j,k,ai,maxa;
 int stop; 
 char sym;

 maxa = proA->Naa + proB->Naa;

 i = si; j=sj; ai = 1; stop = 0;
 while (((i<=proA->Naa)||(j<=proB->Naa))&&(stop==0))
  {
      if (path[i][j]=='v')
      do  
      { a->nalA[ai] = i; a->nalB[ai] = -1; ++i; ++ai;  }
        while ((vpath[i-1][j]=='e')&&(i<=proA->Naa)&&(ai<=maxa));
 
   else if (path[i][j]=='h')
      do 
      { a->nalA[ai] = -1; a->nalB[ai] = j; ++j; ++ai;  } 
        while ((hpath[i][j-1]=='e')&&(j<=proB->Naa)&&(ai<=maxa));
  
  else if (path[i][j]=='d')
         { a->nalA[ai] = i; a->nalB[ai] =j; ++i; ++j; ++ai; }
  else if (path[i][j]=='0') stop = 1; 
  else if (i > proA->Naa) {printf("WOOPS proA Naa over\n"); exit(1);} 
  else if (j > proB->Naa) {printf("WOOPS proB Naa over\n"); exit(1);} 
 }

 a->Nali = ai-1;

} /* end of Trace_Path() */



float Calculate_AlignScore(pA,pB,type,a)
 struct PROTEIN *pA,*pB;
 char type;
 struct AlignNode *a;
{ int i,k;
  char etypeA, etypeB, btypeA, btypeB; 
  float Score;
  struct AlignNode *sn;
  
       if (type == 'B'){ etypeA = etypeB = btypeA = btypeB = '0';}  
  else if (type == 'M')
    { if (pA->Naa > pB->Naa) { etypeA = btypeA = '0'; etypeB = btypeB = '-';}  
      if (pB->Naa > pA->Naa) { etypeA = btypeA = '-'; etypeB = btypeB = '0';}  }
  else if (type == 'L')  etypeA = etypeB = '0';
  else etypeA = etypeB = btypeA = btypeB = '-';

  /*
  printf("Itype %c eA %c eB %c iA %c iB %c\n",
    type,etypeA,etypeB,btypeA,btypeB);  
  */ 
  Score = 0.0;
  for (i=1;i<=a->Nali;++i)
   {
    if ((a->nalA[i]!= -1)&&(a->nalB[i]!= -1)) 
         Score += ResMatchScore(pA,a->nalA[i],pB,a->nalB[i],sn,'-'); 

    if (i>1)
     { if ((a->nalA[i]== -1)&&(a->nalA[i-1]!= -1)) Score += PAR.GapOpenE;
       if ((a->nalA[i]== -1)&&(a->nalA[i-1]== -1)) Score += PAR.GapExtE;
       if ((a->nalB[i]== -1)&&(a->nalB[i-1]!= -1)) Score += PAR.GapOpenE;
       if ((a->nalB[i]== -1)&&(a->nalB[i-1]== -1)) Score += PAR.GapExtE; }
    if (i==1)
     { if ((a->nalA[i]== -1)||(a->nalB[i]== -1)) Score += PAR.GapOpenE; }
  }

  /***** Correction of Boundary ****/

  if (btypeA=='0')
   { i = 1;  while ((i<=a->Nali)&&(a->nalB[i]== -1))  ++i; 
     if (i!=1) Score -= PAR.GapExtE * (i-2) + PAR.GapOpenE; }
   
  if (etypeA=='0')
   {  i = a->Nali; while ((i>=1)&&(a->nalB[i]== -1)) --i;
      if (i!=a->Nali) Score -= PAR.GapExtE * (a->Nali-i-1) + PAR.GapOpenE; }

  if (btypeB=='0')
   { i = 1;  while ((i<=a->Nali)&&(a->nalA[i]== -1))  ++i; 
     if (i!=1) Score -= PAR.GapExtE * (i-2) + PAR.GapOpenE; }
   
  if (etypeB=='0')
    { i = a->Nali; while ((i>=1)&&(a->nalA[i]== -1)) --i;
      if (i!=a->Nali) Score -= PAR.GapExtE * (a->Nali-i-1) + PAR.GapOpenE; }

  return(Score);

} /* end of Calculate_AlignScore() */

